//
// Created by Swen van Zanten on 2019-02-12.
// Copyright (c) 2019 Verge Currency. All rights reserved.
//

import Foundation
import BitcoinKit
import CryptoSwift

class Credentials {

    enum CredentialsError: Error {
        case invalidDeriver(value: String)
        case invalidWidHex(id: String)
    }

    var seed: Data = Data()
    var network: Network = .mainnetXVG

    init(mnemonic: [String], passphrase: String, network: Network = .mainnetXVG) {
        self.seed = Mnemonic.seed(mnemonic: mnemonic, passphrase: passphrase)
        self.network = network
    }

    func reset(mnemonic: [String], passphrase: String, network: Network = .mainnetXVG) {
        self.seed = Mnemonic.seed(mnemonic: mnemonic, passphrase: passphrase)
        self.network = network
    }

    var privateKey: HDPrivateKey {
        return HDPrivateKey(seed: seed, network: network)
    }

    var walletPrivateKey: HDPrivateKey {
        return try! privateKey.derived(at: 0, hardened: true)
    }

    var requestPrivateKey: HDPrivateKey {
        return try! privateKey.derived(at: 1, hardened: true).derived(at: 0)
    }

    var bip44PrivateKey: HDPrivateKey {
        return try! privateKey
            .derived(at: 44, hardened: true)
            .derived(at: 0, hardened: true)
            .derived(at: 0, hardened: true)
    }

    var publicKey: HDPublicKey {
        return bip44PrivateKey.extendedPublicKey()
    }

    var personalEncryptingKey: String {
        let data = Crypto.sha256(requestPrivateKey.privateKey().data)
        let key = "personalKey".data(using: .utf8)!

        let b2 = try! HMAC(key: key.bytes, variant: .sha256).authenticate(data.bytes)

        return Data(b2[0..<16]).base64EncodedString()
    }

    var sharedEncryptingKey: String {
        let sha256Data = walletPrivateKey.privateKey().data.sha256()

        return sha256Data[0..<16].base64EncodedString()
    }

    func privateKeyBy(path: String, privateKey: HDPrivateKey) throws -> PrivateKey {
        var key = privateKey
        for deriver in path.replacingOccurrences(of: "m/", with: "").split(separator: "/") {
            guard let deriverInt32 = UInt32(deriver) else {
                throw CredentialsError.invalidDeriver(value: String(deriver))
            }

            key = try key.derived(at: deriverInt32)
        }

        return key.privateKey()
    }

    func buildSecret(walletId: String) throws -> String {
        guard let widHex = Data(fromHex: walletId.replacingOccurrences(of: "-", with: "")) else {
            throw CredentialsError.invalidWidHex(id: walletId)
        }

        let widBase58 = Base58.encode(widHex).padding(toLength: 22, withPad: "0", startingAt: 0)

        return "\(widBase58)\(self.privateKey.privateKey().toWIF())Lxvg"
    }
}
